import functools
import inspect
from types import AsyncGeneratorType

from fishhook import hook


"""
This script implements decorators, yield_to_sync and yield_to_async, which allow you
to leverage yielding in order to implement sync and async versions of API in a DRY-fashion.
This seems better than the existing alternatives that have either maintenance or compute overhead:
    - code duplication (see unasync or https://spwoodcock.dev/blog/2025-02-python-dry-async/)
    - write a sync version, and run it async via a separate thread
    - write an async version, and use asyncio.run

This works by putting a yield anwyhere you would expect to place an await. The yield then delegates the yielded expression to the aforementioned decorators. In the synchronous case, the expression is just left untounched (hence exhaust_by_identity). In the asynchronous case, the expression is yielded to an async function where we can succesfully apply an await if we get a coroutine and return execution back to the yield. This is essentially exactly how an event loop works, so in effect 'yield-based' approach can be seen as a general form of writing functions, which can specialise to either synchronous functions or asynchronous functions via use of the decorator.

For an example use-case, see chaiverse/database/redis_database.py.

Caveats:
    - Can complicate tracebacks/debugging
    - You cannot write functions like

      def foo(bar):
          return yield bar

      as this is invalid syntax in Python. You must assign the yield expression to a variable, and return that, as follows:

      def foo(bar):
        baz = yield bar
        return baz
    - 'async for' loops work using the usual for loop syntax, provided the for
       loop includes the following code:

       from chaiverse.lib.async_tools import YieldStop
       for i in possibly_async_generator():
           i = yield i
           if i == YieldStop; break

        The first line 'i = yield i' makes sense: at some point you must await the result of the generator if it is async. The second line is a bit of hack, and comes from the fact that it seems not possible
        to know whether an async generator has finished yielding until after the fact. This causes an extra value to be yielded to the for loop when exhausting. So we explicitly indicate the loop needs to break.

TODO: To the best of my knowledge, this approach is novel. The package https://github.com/dry-python/returns does something similar using Futures, but it seems messier and less intuitive. We could make this an open-source package.
"""


def is_generator(gen):
    return inspect.isgenerator(gen)


def is_coroutine(coro):
    return inspect.iscoroutine(coro)


class YieldStop:
    pass


@hook(AsyncGeneratorType)
def __iter__(self):
    exhausted = False
    async def advance_async_iterator():
        try:
            out = await anext(self)
        except StopAsyncIteration:
            nonlocal exhausted
            exhausted = True
            out = YieldStop
        return out
    while not exhausted:
        yield advance_async_iterator()



def exhaust_by_identity(gen):
    try:
        out = next(gen)
        while True:
            try:
                out = exhaust_by_identity(out) if is_generator(out) else out
                out = gen.send(out)
            except Exception as ex:
                if isinstance(ex, StopIteration): raise ex
                out = gen.throw(ex)
    except StopIteration as ex:
        out = ex.value
    return out


def yield_to_sync(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        gen = func(*args, **kwargs)
        return exhaust_by_identity(gen)
    return wrapper


async def exhaust_by_await(gen):
    try:
        out = next(gen)
        while True:
            try:
                out = await exhaust_by_await(out) if is_generator(out) else out
                out = await out if is_coroutine(out) else out
                out = gen.send(out)
            except Exception as ex:
                if isinstance(ex, StopIteration): raise ex
                out = gen.throw(ex)
    except StopIteration as ex:
        out = ex.value
    return out


def yield_to_async(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        gen = await func(*args, **kwargs)
        return await exhaust_by_await(gen)
    return wrapper



import asyncio
import functools
import os
from typing import Optional

from flatten_dict import flatten
import orjson
from redis import exceptions, Redis
from redis.asyncio import Redis as AsyncRedis, BlockingConnectionPool
from redis.commands.json.path import Path
from redis.commands.search.field import TextField
from redis.commands.search.index_definition import IndexDefinition, IndexType
from redis.commands.search import Search
from redis.exceptions import ResponseError

from chaiverse.database.inferno_database_adapter import _InfernoDatabaseAdapter
from chaiverse.lib.async_tools import yield_to_async, yield_to_sync, YieldStop


class JSONDecoder():
    def decode(obj):
        return orjson.loads(obj)


class _AbstractRedisDatabase(_InfernoDatabaseAdapter):
    """
       Redis database where each record is aggregated under a single Redis key,
       which corresponds to the first level keys of the Python dictionary.
    """
    def __init__(self, url: str, port: int, password: str, ttl: int = 0):
        self.url = url
        self.port = port
        self.password = password
        self.ttl = ttl
        self.client = Redis(
            host=url, port=port, username="default", password=password, decode_responses=True
        )
        self.json_client = self.client.json(decoder=JSONDecoder)

    def set(self, path: str, value: dict):
        yield self._set_json(path, value)

    def get(self, path: str, shallow: bool = False):
        record = yield self._get_json(path)
        record = record[0] if record else None
        return record

    def is_in_database(self, path: str):
        value = yield self.get(path, shallow=True)
        return value is not None

    def update(self, path: str, record: dict):
        yield self._merge_json(path, record)

    def multi_update(self, path: str, record: dict):
        record = flatten(record, reducer="path", max_flatten_depth=2)
        for key, value in record.items():
            key = f"{path}/{key}"
            yield self._merge_json(key, value)

    def remove(self, path: str):
        key, path = self._get_key_and_path(path)
        yield self.json_client.delete(key, path)

    def where(self, path: str, **kwargs):
        assert len(kwargs) == 1, "Searching by only one field value is currently supported!"
        field = list(kwargs.keys())[0]
        value = list(kwargs.values())[0]
        results = yield self._get_filtered_json(path, field, value)
        return results

    def _check_health(self):
        health = yield self.client.ping()
        return health

    def _get_json(self, path):
        key, path = self._get_key_and_path(path)
        record = yield self.json_client.get(key, path)
        return record

    def _get_filtered_json(self, path, field, value):
        key, _ = self._get_key_and_path(path)
        query = f"$..[?(@.{field} == '{value}')]"
        results = yield self.json_client.get(key, query)
        return results

    def _set_json(self, path, record):
        key, path = self._get_key_and_path(path)
        yield self._set_json_with_retry(key, path, record)

    def _merge_json(self, path, record):
        key, path = self._get_key_and_path(path)
        # TODO: We don't rely on Redis' merge rules, as it is not the same as
        #       Firebase's. So to maintain a consistent behaviour, we mimic
        #       Firebase's merge rules, which does not update nested dicts.
        #       Redis' rule is more sane, though, so we should migrate to that
        #       as the default.
        current_value = yield self.json_client.get(key, path)
        current_value = current_value[0] if current_value else None
        record = {**current_value, **record} if current_value else record
        yield self._set_json_with_retry(key, path, record)

    def _set_json_with_retry(self, key, path, record):
        record = _ignore_null_values(record)
        try:
            yield self.json_client.set(key, path, record)
        except exceptions.ResponseError:
            yield self.json_client.set(key, "$", {})
            yield self.json_client.set(key, path, record)
        yield self._expire_key(key)

    def _expire_key(self, key):
        if self.ttl > 0 and self.ttl < 1:
            ttl_ms = int(self.ttl * 1000)
            yield self.client.pexpire(key, ttl_ms)
        elif self.ttl > 0:
            yield self.client.expire(key, self.ttl)

    def _get_key_and_path(self, path):
        path = path.lstrip("/")
        split_path = path.split("/", 1)
        key = split_path[0]
        key = _clean_key(key)
        if len(split_path) > 1:
            path = split_path[1].replace("/", ".")
            path = f"$.{path}"
        else:
            path = "$"
        return key, path


class _RedisDatabase(_AbstractRedisDatabase):
    @yield_to_sync
    def set(self, path: str, value: dict):
        return super().set(path, value)

    @yield_to_sync
    def get(self, path: str, shallow: bool = False):
        return super().get(path, shallow)

    @yield_to_sync
    def update(self, path: str, record: dict):
        return super().update(path, record)

    @yield_to_sync
    def multi_update(self, path: str, record: dict):
        return super().multi_update(path, record)

    @yield_to_sync
    def remove(self, path: str):
        return super().remove(path)

    @yield_to_sync
    def where(self, path: str, **kwargs):
        return super().where(path, **kwargs)

    @yield_to_sync
    def _check_health(self):
        return super()._check_health()


class _AsyncRedisDatabase(_AbstractRedisDatabase):
    def __init__(self, url: str, port: int, password: str, ttl: int = 0):
        self.url = url
        self.port = port
        self.password = password
        self.ttl = ttl
        max_connections = int(os.environ.get("REDIS_MAX_CONNECTIONS", 100))
        pool = BlockingConnectionPool(
            host=url,
            port=port,
            username="default",
            password=password,
            decode_responses=True,
            max_connections=max_connections,
            timeout=0.5
        )
        self.client = AsyncRedis.from_pool(pool)
        self.json_client = self.client.json(decoder=JSONDecoder)

    @yield_to_async
    async def set(self, path: str, value: dict):
        return super().set(path, value)

    @yield_to_async
    async def get(self, path: str, shallow: bool = False):
        return super().get(path, shallow)

    @yield_to_async
    async def update(self, path: str, record: dict):
        return super().update(path, record)

    @yield_to_async
    async def multi_update(self, path: str, record: dict):
        return super().multi_update(path, record)

    @yield_to_async
    async def remove(self, path: str):
        return super().remove(path)

    @yield_to_async
    async def where(self, path: str, **kwargs):
        return super().where(path, **kwargs)

    @yield_to_async
    async def _check_health(self):
        return super()._check_health()


class _MultiRecordAbstractRedisDatabase(_AbstractRedisDatabase):
    """
       Variant of Redis database, where each document is stored at the Redis
       key that equals the path. This is useful for situations where TTLs need
       to be set for each record.
    """
    def _get_key_and_path(self, path):
        path = path.lstrip("/")
        key = _clean_key(path)
        path = "$"
        return key, path

    def get(self, path: str, shallow: bool = False):
        path = _clean_key(path)
        func = super().get if path.count("/") >= 1 else self._multi_get
        out = yield func(path)
        return out

    def where(self, *args, **kwargs):
        raise NotImplementedError

    def _multi_get(self, path: str):
        result = {}
        for key in self.client.scan_iter(f"{path}*"):
            key = yield key
            if key == YieldStop: break
            record = yield self._get_json(key)
            record = record[0] if record else None
            dict_key = key.replace(path, "")
            dict_key = _clean_key(dict_key)
            result[dict_key] = record
        if len(result) == 1:
            result = list(result.values())[0]
        return result


class _MultiRecordRedisDatabase(_MultiRecordAbstractRedisDatabase, _RedisDatabase):
    """
       Variant of Redis database, where each document is stored at the Redis
       key that equals the path. This is useful for situations where TTLs need
       to be set for each record.
    """
    @yield_to_sync
    def get(self, path: str, shallow: bool = False):
        return super().get(path, shallow)


class _AsyncMultiRecordRedisDatabase(_MultiRecordAbstractRedisDatabase, _AsyncRedisDatabase):
    """
       Variant of Redis database, where each document is stored at the Redis
       key that equals the path. This is useful for situations where TTLs need
       to be set for each record.
    """
    @yield_to_async
    async def get(self, path: str, shallow: bool = False):
        return super().get(path, shallow)


def _clean_key(key):
    key = key.replace("//", "/")
    key = key.lstrip("/")
    return key


def _ignore_null_values(d):
    return {k: v for k,v in d.items() if v is not None}



from cachetools import cached
import inspect
from typing import List, Dict

from pydantic import BaseModel


@cached(cache={})
def get_fields_in_schema(cls: BaseModel) -> List[str]:
    fields = list(cls.__fields__.keys()) + _get_class_properties(cls)
    fields = _remove_protected_fields(fields)
    return fields


def _remove_protected_fields(fields):
    fields = [field for field in fields if not field.startswith('_')]
    fields = [field for field in fields if field != "model_fields_set"]
    fields = [field for field in fields if field != "model_extra"]
    return fields


def _get_class_properties(cls):
    properties = [
        member_name for member_name, member_object in inspect.getmembers(cls)
        if inspect.isdatadescriptor(member_object)
    ]
    return properties


class UnionParser:
    '''
    It took quite sometime to find this approach, the parser can be used across all pydantic versions.
    Synopsys:    parse a dict into a union of pydantic models and auto pick the model that can fit.
    Usage:       UnionParser[ModelA|ModelB].parse(<dict>)
    Description: There is no built-in method to parse a dict into a union of pydantic models. I figure 
    out it is the easiest to parse it into a container class that has a property of this union type.
    pydantic already supported this type of container class parsing. I searched all over internet and
    this approach is not mentioned anywhere. Either upgrading to pydantic_v2 or use a __root__ field
    was recommended in other places.
    Conceivably, this method can be used to parse any type into a pydantic model, but parse_obj_as
    from pydantic probably should be used for all cases that pydantic can natievly support parsing.
    '''
    def __class_getitem__(cls, union_type):
        class _UnionContainer(BaseModel):
            value: union_type

            @classmethod
            def parse(cls, value: Dict):
                return _UnionContainer(value=value).value

        return _UnionContainer



from .inferno_database_adapter import _InfernoDatabaseAdapter
from .firebase_database import _FirebaseDatabase, serialise_input, _get_disallowed_firebase_keys, sanitise_database_keys, sanitise_database_string
from .redis_database import _AsyncRedisDatabase, _RedisDatabase, _MultiRecordRedisDatabase, _AsyncMultiRecordRedisDatabase
from .mock_database import AsyncMockDatabase, MockDatabase
